/*
 * @author     Ekawat (ice) Homsirikamol
 * @version    1.0
 * @copyright  Copyright (c) 2017 Cryptographic Engineering Research Group
 *             ECE Department, George Mason University Fairfax, VA, U.S.A.
 *             All rights Reserved.
 * @license    This project is released under the GNU Public License.
 *             The license and distribution terms for this file may be
 *             found in the file LICENSE in this distribution or at
 *             http://www.gnu.org/licenses/gpl-3.0.txt
 * @note       This implementation is derived in part from the reference implementation
 *             included in SUPERCOP-20171020, with the list of co-authors listed in ref/designers
 *             
 *             This is publicly available encryption source code that falls
 *             under the License Exception TSU (Technology and software-
 *             unrestricted)
 */

#if !__SYNTHESIS__
#include <iostream>
#endif

#include "ap_int.h"
#include "hls_stream.h"
using namespace hls;
#include "hls_interface.h"
#include "hls_utils.h"
#include "hls_aes.h"


#define GCM_ROUND 8

#if !__SYNTHESIS__
#define DBG 1
#define DBG_IN 1
#define DBG_OUT 1
#define DBG_ROUND 0
#endif

enum states_t {S_HASH_KEY, S_EK01, S_DATA, S_TAG};


void mult128x16(ap_uint<8> new_vv[16], ap_uint<8> new_zz[16], ap_uint<8> vv[16], ap_uint<8> zz[16], ap_uint<16> hh)
{
    int i;
    ap_uint<8> zztmp[16], vvtmp[16];
    ap_uint<8> zztmp2[16], vvtmp2[16];
#pragma HLS ARRAY_RESHAPE variable=zz      complete dim=1
#pragma HLS ARRAY_RESHAPE variable=vv      complete dim=1
#pragma HLS ARRAY_RESHAPE variable=new_vv  complete dim=1
#pragma HLS ARRAY_RESHAPE variable=new_zz  complete dim=1
#pragma HLS ARRAY_RESHAPE variable=zztmp   complete dim=1
#pragma HLS ARRAY_RESHAPE variable=zztmp2  complete dim=1
#pragma HLS ARRAY_RESHAPE variable=vvtmp   complete dim=1
#pragma HLS ARRAY_RESHAPE variable=vvtmp2  complete dim=1

    copy(zztmp, zz);
    copy(vvtmp, vv);
    for (i=0; i<16; i++) {
#pragma HLS UNROLL
        if (((hh >> (15-i)) & 0x1) == 1) 
            xor2(zztmp2, zztmp, vvtmp);
        else 
            copy(zztmp2, zztmp);
        shf_right1(vvtmp2, vvtmp);
        if ((vvtmp[15] & 0x1) == 1) 
            vvtmp2[0] ^= 0xE1;
        copy(vvtmp, vvtmp2);
        copy(zztmp, zztmp2);
    }
    copy(new_vv, vvtmp2);
    copy(new_zz, zztmp2);
}

void hls_aes_gcm (
    BDO_STREAM& bdo_bus,
    bool *msg_auth_valid,
    BDI_STREAM& bdi_bus,
    KEY_STREAM& key_bus,
    bool decrypt,
    bool key_update
    )
{
    /* Top-Level Interface */
#pragma HLS DATA_PACK variable=bdo_bus
#pragma HLS DATA_PACK variable=bdi_bus
#pragma HLS DATA_PACK variable=key_bus
#pragma HLS INTERFACE port=bdo_bus axis
#pragma HLS INTERFACE port=bdi_bus axis
#pragma HLS INTERFACE port=key_bus axis
#pragma HLS INTERFACE ap_vld  port=msg_auth_valid

    /* Static variable */
    static array_t key_data;
    static array_t hkey;

#pragma HLS ARRAY_RESHAPE variable=key_data complete dim=1
#pragma HLS ARRAY_RESHAPE variable=hkey     complete dim=1

    /* Semi-static variables */ 
    states_t state;
    ap_uint<WORDSIZE> iv[12];
    array_t ek0;
    array_t tag;
    array_t aes_do;

#pragma HLS ARRAY_RESHAPE variable=iv       complete dim=1
#pragma HLS ARRAY_RESHAPE variable=ek0      complete dim=1
#pragma HLS ARRAY_RESHAPE variable=tag      complete dim=1
#pragma HLS ARRAY_RESHAPE variable=aes_do complete dim=1

    /* Interface related temporary variables */
    public_bus bdi;
    output_bus bdo;
    secret_bus key;
    array_t bdo_data;
    array_t bdi_data;

#pragma HLS ARRAY_RESHAPE variable=bdo_data complete dim=1
#pragma HLS ARRAY_RESHAPE variable=bdi_data complete dim=1

    /* Temporary variables */
    ap_uint<16> hh;
    ap_uint<4> round;
    array_t to_aes;
    array_t to_mult;
    array_t from_bdi_xor;
    array_t to_bdo;
    array_t vv;
    array_t zz;
    array_t add;
    array_t rk;
    array_t add_enc;
    array_t shift;
    array_t from_aes;
    array_t new_rk;
    array_t new_vv;
    array_t new_zz;
    ap_uint<BLOCKSIZE> hkeyBlock;

#pragma HLS ARRAY_RESHAPE variable=to_aes complete dim=1
#pragma HLS ARRAY_RESHAPE variable=to_mult complete dim=1
#pragma HLS ARRAY_RESHAPE variable=from_bdi_xor complete dim=1
#pragma HLS ARRAY_RESHAPE variable=to_bdo complete dim=1
#pragma HLS ARRAY_RESHAPE variable=vv complete dim=1
#pragma HLS ARRAY_RESHAPE variable=zz complete dim=1
#pragma HLS ARRAY_RESHAPE variable=add complete dim=1
#pragma HLS ARRAY_RESHAPE variable=rk complete dim=1
#pragma HLS ARRAY_RESHAPE variable=add_enc complete dim=1
#pragma HLS ARRAY_RESHAPE variable=shift complete dim=1
#pragma HLS ARRAY_RESHAPE variable=from_aes complete dim=1
#pragma HLS ARRAY_RESHAPE variable=new_rk complete dim=1
#pragma HLS ARRAY_RESHAPE variable=new_vv complete dim=1
#pragma HLS ARRAY_RESHAPE variable=new_zz complete dim=1

    int i;
    
    /* Initialization */
    size_t loop = 1;
    bool is_first = 1;
    ap_uint<32> ctr = 1;
    ap_uint<32> len_a = 0;
    ap_uint<32> len_d = 0;
    clear(ek0);
    clear(tag);
    bdo.size = 0;
#if !__SYNTHESIS__
    size_t dbg_loop = 0;
#endif

    // Key initialization (if needed)
    if (key_update == 1) {
        key_bus.read(key); reshape_block(key_data, key.data);
        clear(hkey);
        state = S_HASH_KEY;
    } else {
        state = S_EK01;
    }
    
    // Read IV
    bdi_bus.read(bdi);
    reshape_block(bdi_data, bdi.data);
    for (i=0; i<12; i++) {
    #pragma HLS UNROLL
        iv[i] = bdi_data[i];
    }

    while (loop) {
#if !__SYNTHESIS__
        // Add this clause to prevent infinite loop during debugging
        if (dbg_loop > 300) return;
        else dbg_loop++;
#endif
#if DBG
        std::cout << "Current state = " << state << std::endl;
        std::cout << "\tctr = " << ctr << std::endl;
#endif
        // PreProcessing
        switch (state) {
        case S_HASH_KEY:
            clear(to_aes);
            clear(to_mult);
            round = AES_ROUND;
            break;
        case S_EK01:
            for (i=0; i<12; i++) {
            #pragma HLS UNROLL
                to_aes[i] = iv[i];
            }
            to_aes[12] = (((ap_uint<8>) (ctr >> 24)) & 0xff);
            to_aes[13] = (((ap_uint<8>) (ctr >> 16)) & 0xff);
            to_aes[14] = (((ap_uint<8>) (ctr >>  8)) & 0xff);
            to_aes[15] = (((ap_uint<8>) (ctr >>  0)) & 0xff);
            clear(to_mult);
            round = AES_ROUND;
            break;
        case S_DATA:
            bdi_bus.read(bdi); reshape_block(bdi_data, bdi.data);
#if DBG
            std::cout << "Reading data" << std::endl;
            std::cout << "\tvbytes: " << std::hex << std::uppercase << std::setw(4) << bdi.valid_bytes << std::endl;
            std::cout << "\tpad   : " << std::hex << std::uppercase << std::setw(4) << bdi.pad_loc << std::endl;
#endif
            if ((bdi.type >> 1)  == TYPE_AD) { // AD
                len_a = len_a + ((ap_uint<8>) bdi.size << 3);
                round = GCM_ROUND;
                copy(to_mult, bdi_data);
            } else { // Data
                len_d = len_d + ((ap_uint<8>) bdi.size << 3);
                if (bdi.eot)
                    round = GCM_ROUND;
                else
                    round = AES_ROUND;
                xor2(from_bdi_xor, bdi_data, aes_do);
#if DBG
                std::cout << "Writing data" << std::endl;
                std::cout << "\tAES: " << std::hex << std::uppercase << std::setw(16) << reshape_array(aes_do) << std::endl;
                std::cout << "\tBDI: " << std::hex << std::uppercase << std::setw(16) << reshape_array(bdi_data) << std::endl;
                std::cout << "\tBDO: " << std::hex << std::uppercase << std::setw(16) << reshape_array(from_bdi_xor) << std::endl;

#endif
                and_valid(to_bdo, from_bdi_xor, bdi.valid_bytes);

                // Mult input
                if (decrypt) 
                    copy(to_mult, bdi_data);
                else
                    copy(to_mult, to_bdo);
                
                // Write output
                bdo.data = reshape_array(to_bdo);
                bdo_bus.write(bdo);
            }
            for (i=0; i<12; i++) {
            #pragma HLS UNROLL
                to_aes[i] = iv[i];
            }
            to_aes[12] = (((ap_uint<8>) (ctr >> 24)) & 0xff);
            to_aes[13] = (((ap_uint<8>) (ctr >> 16)) & 0xff);
            to_aes[14] = (((ap_uint<8>) (ctr >>  8)) & 0xff);
            to_aes[15] = (((ap_uint<8>) (ctr >>  0)) & 0xff);
            break;
        default: //S_TAG
            for (i=0; i<4; i++) {
            #pragma HLS UNROLL
                to_mult[i] = 0;
            }
            to_mult[4] = (((ap_uint<8>) (len_a >> 24)) & 0xff);
            to_mult[5] = (((ap_uint<8>) (len_a >> 16)) & 0xff);
            to_mult[6] = (((ap_uint<8>) (len_a >>  8)) & 0xff);
            to_mult[7] = (((ap_uint<8>) (len_a >>  0)) & 0xff);
            for (i=8; i<12; i++) {
            #pragma HLS UNROLL
                to_mult[i] = 0;
            }
            to_mult[12] = (((ap_uint<8>) (len_d >> 24)) & 0xff);
            to_mult[13] = (((ap_uint<8>) (len_d >> 16)) & 0xff);
            to_mult[14] = (((ap_uint<8>) (len_d >>  8)) & 0xff);
            to_mult[15] = (((ap_uint<8>) (len_d >>  0)) & 0xff);
            clear(to_aes);
            round = GCM_ROUND;
        }
        /* Prep GCM */
        hkeyBlock = reshape_array(hkey);
        hh = (ap_uint<16>) ((hkeyBlock >> 112) & 0xFFFF);
        xor2(vv, tag, to_mult);
        clear(zz);

        /* Initial Add round key */
#if DBG_IN
        std::cout << "\tAES  In: " << std::hex << std::uppercase << std::setw(16) << reshape_array(to_aes) << std::endl;
        std::cout << "\tKey  In: " << std::hex << std::uppercase << std::setw(16) << reshape_array(key_data) << std::endl;
        std::cout << "\tmult In: " << std::hex << std::uppercase << std::setw(16) << reshape_array(vv) << std::endl;
#endif

        AddRoundKey(add, to_aes, key_data);
        AESKeyUpdate(rk, key_data, 0);

        /* ========= */
        /* MAIN LOOP */
        /* ========= */
        AES_LOOP: 
        for (i=0; i<round; i++) {
#if DBG_ROUND
            std::cout << "\tAES Round(" << i << "):" << std::hex << std::uppercase << std::setw(16) << reshape_array(add) << std::endl;
            std::cout << "\tKey Round(" << i << "):" << std::hex << std::uppercase << std::setw(16) << reshape_array(rk) << std::endl;
            std::cout << "\tvv Round(" << i << "):" << std::hex << std::uppercase << std::setw(16) << reshape_array(vv) << std::endl;
            std::cout << "\tzz Round(" << i << "):" << std::hex << std::uppercase << std::setw(16) << reshape_array(zz) << std::endl;
#endif
            AESKeyUpdate(new_rk, rk, i+1);
            AESRound(shift, add_enc, add, rk);
            AddRoundKey(from_aes, shift, rk);
            copy(add, add_enc);
            copy(rk, new_rk);


            /* GCM */
            mult128x16(new_vv, new_zz, vv, zz, hh);

            if (i == 0)
                hh = (ap_uint<16>) ((hkeyBlock >> 96) & 0xFFFF);
            else if (i == 1)
                hh = (ap_uint<16>) ((hkeyBlock >> 80) & 0xFFFF);
            else if (i == 2)
                hh = (ap_uint<16>) ((hkeyBlock >> 64) & 0xFFFF);
            else if (i == 3)
                hh = (ap_uint<16>) ((hkeyBlock >> 48) & 0xFFFF);
            else if (i == 4)
                hh = (ap_uint<16>) ((hkeyBlock >> 32) & 0xFFFF);
            else if (i == 5)
                hh = (ap_uint<16>) ((hkeyBlock >> 16) & 0xFFFF);
            else
                hh = (ap_uint<16>) ((hkeyBlock >>  0) & 0xFFFF);

            if (i < 8) {
                copy(vv, new_vv);
                copy(zz, new_zz);
            }
        }

#if DBG_OUT
        std::cout << "\tAES  Out: " << std::hex << std::uppercase << std::setw(16) << reshape_array(from_aes) << std::endl;
        std::cout << "\tMult Out: " << std::hex << std::uppercase << std::setw(16) << reshape_array(zz) << std::endl;
#endif

        // Post Processing
        switch (state) {
        case S_HASH_KEY:
            copy(hkey, from_aes);
            state = S_EK01;
            break;
        case (S_EK01):
            if (ctr == 2)  {
                state = S_DATA;
            } else {
                copy(ek0, from_aes);
                if (bdi.eoi)
                    state = S_TAG;
            }
            copy(aes_do, from_aes);
            ctr++;
            break;
        case S_DATA:
            is_first = 0;
            copy(tag, zz);
#if DBG
            std::cout << "-- type: " << bdi.type << std::endl;
#endif
            if (bdi.type == TYPE_D) {
                copy(aes_do, from_aes);
                ctr++;
            }
            if (bdi.eoi)
                state = S_TAG;
            break;
        default: // S_TAG
            xor2(to_bdo, zz, ek0);
            if (decrypt) {
                bdi_bus.read(bdi); reshape_block(bdi_data, bdi.data);
                *msg_auth_valid = match(bdi_data, to_bdo);
            } else {
                bdo.data = reshape_array(to_bdo);
                bdo_bus.write(bdo);
            }
            return;
        }
    }
}
